module Conjunto
where

-- Definicion del tipo
data Conjunto a = Vacio | Elemento a (Conjunto a) deriving Show

-- Devuelve el Conjunto Vacio
vacio :: Conjunto a
vacio = Vacio

-- Toma un Conjunto y determina si es vacio o no
es_vacio :: Conjunto a -> Bool
es_vacio Vacio = True
es_vacio (Elemento elem conj) = False

-- Toma un Conjunto y devuelve la cantidad de elementos.
-- Nota: esta definicion asume que todos los elementos del 
-- conjunto son distintos.
-- Queda como ejercicio mejorarla para que se consideren
-- los conjuntos con elementos repetidos
elementos :: Conjunto a -> Int
elementos Vacio = 0
elementos (Elemento elem conj) = 1 + (elementos conj)

-- Toma un elemento y un Conjunto y determina si el elemento
-- esta en el conjunto
pertenece :: Eq a => a -> Conjunto a -> Bool
pertenece elem Vacio = False
pertenece elem (Elemento e cj) = (elem == e) || (pertenece elem cj)

-- Toma dos Conjuntos y determina si el primero está incluido
-- en el segundo
incluido :: Eq a => Conjunto a -> Conjunto a -> Bool
incluido Vacio _ = True
incluido (Elemento elem cj1) cj2 = (pertenece elem cj2) && (incluido cj1 cj2)

-- Toma dos Conjuntos y determina si son iguales
-- (es decir, si guardan los mismos elementos)
iguales :: Eq a => Conjunto a -> Conjunto a -> Bool
iguales cj1 cj2 = (incluido cj1 cj2) && (incluido cj2 cj1)

-- Conjuntos de ejemplo
c1,c2,c3,c4 :: Conjunto Int
c1 = Vacio
c2 = Elemento 1 (Elemento 5 (Elemento 6 Vacio))
c3 = Elemento 5 (Elemento 4 (Elemento 1 (Elemento 6 Vacio)))
c4 = Elemento 6 (Elemento 4 (Elemento 5 (Elemento 1 Vacio)))

-- Intentaremos hacer que (Conjunto a) cumplan (sean parte de) Eq y Ord,
-- para definiciones arbitrarias de == y de <= (que pueden cambiarse a
-- piacere)

instance Eq a => Eq (Conjunto a) where
    cj1 == cj2 = iguales cj1 cj2

instance Ord a => Ord (Conjunto a) where
    cj1 <= cj2 = incluido cj1 cj2
